#include "common.c"
#include "crypto_box.h"
#include <sys/types.h>
#include <sys/time.h>

#define uint64 unsigned long long //typedef unsigned long long uint64;

struct tai {
  uint64 x;
};
struct taia {
  struct tai sec;
  unsigned long nano; /* 0...999999999 */
  unsigned long atto; /* 0...999999999 */
};

struct qt_proto_data_nacltai {
	unsigned char cenonce[crypto_box_NONCEBYTES], cdnonce[crypto_box_NONCEBYTES];
	unsigned char cbefore[crypto_box_BEFORENMBYTES];
	struct taia cdtaip, cdtaic;
};

#define noncelength 16
#define nonceoffset (crypto_box_NONCEBYTES - noncelength)
/*static unsigned char cbefore[crypto_box_BEFORENMBYTES];
static unsigned char buffer1[MAX_PACKET_LEN+crypto_box_ZEROBYTES], buffer2[MAX_PACKET_LEN+crypto_box_ZEROBYTES];
static const unsigned char* buffer1offset = buffer1 + crypto_box_ZEROBYTES;
static const unsigned char* buffer2offset = buffer2 + crypto_box_BOXZEROBYTES - noncelength;*/
static const int overhead                 = crypto_box_BOXZEROBYTES + noncelength;

void tai_pack(char *s, struct tai *t) {
  uint64 x;
  x = t->x;
  s[7] = x & 255; x >>= 8;
  s[6] = x & 255; x >>= 8;
  s[5] = x & 255; x >>= 8;
  s[4] = x & 255; x >>= 8;
  s[3] = x & 255; x >>= 8;
  s[2] = x & 255; x >>= 8;
  s[1] = x & 255; x >>= 8;
  s[0] = x;
}
void tai_unpack(char *s, struct tai *t) {
  uint64 x;
  x = (unsigned char) s[0];
  x <<= 8; x += (unsigned char) s[1];
  x <<= 8; x += (unsigned char) s[2];
  x <<= 8; x += (unsigned char) s[3];
  x <<= 8; x += (unsigned char) s[4];
  x <<= 8; x += (unsigned char) s[5];
  x <<= 8; x += (unsigned char) s[6];
  x <<= 8; x += (unsigned char) s[7];
  t->x = x;
}
void taia_pack(char *s, struct taia *t) {
  unsigned long x;
  tai_pack(s,&t->sec);
  s += 8;
  x = t->atto;
  s[7] = x & 255; x >>= 8;
  s[6] = x & 255; x >>= 8;
  s[5] = x & 255; x >>= 8;
  s[4] = x;
  x = t->nano;
  s[3] = x & 255; x >>= 8;
  s[2] = x & 255; x >>= 8;
  s[1] = x & 255; x >>= 8;
  s[0] = x;
} 
void taia_unpack(char *s, struct taia *t) {
  unsigned long x;
  tai_unpack(s,&t->sec);
  s += 8;
  x = (unsigned char) s[4];
  x <<= 8; x += (unsigned char) s[5];
  x <<= 8; x += (unsigned char) s[6];
  x <<= 8; x += (unsigned char) s[7];
  t->atto = x;
  x = (unsigned char) s[0];
  x <<= 8; x += (unsigned char) s[1];
  x <<= 8; x += (unsigned char) s[2];
  x <<= 8; x += (unsigned char) s[3];
  t->nano = x;
}

void taia_now(struct taia *t) {
  struct timeval now;
  gettimeofday(&now,(struct timezone *) 0);
  t->sec.x = 4611686018427387914ULL + (uint64) now.tv_sec;
  t->nano = 1000 * now.tv_usec + 500;
  t->atto = 0;
}

extern crypto_scalarmult_curve25519_base(unsigned char *pk, unsigned char *sk);

static int encode(struct qtsession* sess, char* raw, char* enc, int len) {
//	fprintf(stderr, "Encoding packet of %d bytes from %d to %d\n", len, raw, enc);
	struct qt_proto_data_nacltai* d = (struct qt_proto_data_nacltai*)sess->protocol_data;
	memset(raw, 0, crypto_box_ZEROBYTES);
	taia_now(&d->cdtaic);
	taia_pack(d->cenonce + nonceoffset, &(d->cdtaic));
	if (crypto_box_afternm(enc, raw, len + crypto_box_ZEROBYTES, d->cenonce, d->cbefore)) return errorexit("Crypto failed");
	memcpy((void*)(enc + crypto_box_BOXZEROBYTES - noncelength), d->cenonce + nonceoffset, noncelength);
	len += overhead;
//	fprintf(stderr, "Encoded packet of %d bytes from %d to %d\n", len, raw, enc);
	return len;
}

static int decode(struct qtsession* sess, char* enc, char* raw, int len) {
//	fprintf(stderr, "Decoding packet of %d bytes from %d to %d\n", len, enc, raw);
	struct qt_proto_data_nacltai* d = (struct qt_proto_data_nacltai*)sess->protocol_data;
	int i;
	if (len < overhead) {
		fprintf(stderr, "Short packet received: %d\n", len);
		return 0;
	}
	len -= overhead;
	taia_unpack((char*)(enc + crypto_box_BOXZEROBYTES - noncelength), &d->cdtaic);
	if (d->cdtaic.sec.x <= d->cdtaip.sec.x && d->cdtaic.nano <= d->cdtaip.nano && d->cdtaic.atto <= d->cdtaip.atto) { 
		fprintf(stderr, "Timestamp going back, ignoring packet\n");
		return 0;
	}
	memcpy(d->cdnonce + nonceoffset, enc + crypto_box_BOXZEROBYTES - noncelength, noncelength);
	memset(enc, 0, crypto_box_BOXZEROBYTES);
	if (i = crypto_box_open_afternm(raw, enc, len + crypto_box_ZEROBYTES, d->cdnonce, d->cbefore)) {
		fprintf(stderr, "Decryption failed len=%d result=%d\n", len, i);
		return 0;
	}
	d->cdtaip = d->cdtaic;
//	fprintf(stderr, "Decoded packet of %d bytes from %d to %d\n", len, enc, raw);
	return len;
}

static int init(struct qtsession* sess) {
	struct qt_proto_data_nacltai* d = (struct qt_proto_data_nacltai*)sess->protocol_data;
	char* envval;
	printf("Initializing cryptography...\n");
	unsigned char cownpublickey[crypto_box_PUBLICKEYBYTES], cpublickey[crypto_box_PUBLICKEYBYTES], csecretkey[crypto_box_SECRETKEYBYTES];
	if (!(envval = getconf("PUBLIC_KEY"))) return errorexit("Missing PUBLIC_KEY");
	if (strlen(envval) != 2*crypto_box_PUBLICKEYBYTES) return errorexit("PUBLIC_KEY length");
	hex2bin(cpublickey, envval, crypto_box_PUBLICKEYBYTES);
	if (!(envval = getconf("PRIVATE_KEY"))) return errorexit("Missing PRIVATE_KEY");
	if (strlen(envval) != 2*crypto_box_PUBLICKEYBYTES) return errorexit("PRIVATE_KEY length");
	hex2bin(csecretkey, envval, crypto_box_SECRETKEYBYTES);
	crypto_box_beforenm(d->cbefore, cpublickey, csecretkey);

	memset(d->cenonce, 0, crypto_box_NONCEBYTES);
	memset(d->cdnonce, 0, crypto_box_NONCEBYTES);

	const unsigned char base[32] = {9};
	crypto_scalarmult(cownpublickey, csecretkey, base);

	if (envval = getenv("TIME_WINDOW")) {
		taia_now(&d->cdtaip);
		d->cdtaip.sec.x -= atol(envval);
	} else {
		fprintf(stderr, "Warning: TIME_WINDOW not set, risking an initial replay attack\n");
	}
	if (envval = getenv("ROLE")) {
		d->cenonce[nonceoffset-1] = atoi(envval) ? 1 : 0;
	} else {
		d->cenonce[nonceoffset-1] = memcmp(cownpublickey, cpublickey, crypto_box_PUBLICKEYBYTES) > 0 ? 1 : 0;
	}
	d->cdnonce[nonceoffset-1] = d->cenonce[nonceoffset-1] ? 0 : 1;
	return 0;
}

#ifdef COMBINED_BINARY
	int tunmain_nacltai() {
#else
	int tunmain() {
#endif
	struct qtproto p = {
		1,
		MAX_PACKET_LEN + crypto_box_ZEROBYTES,
		MAX_PACKET_LEN + crypto_box_ZEROBYTES,
		crypto_box_ZEROBYTES,
		crypto_box_BOXZEROBYTES - noncelength,
		encode,
		decode,
		init,
		sizeof(struct qt_proto_data_nacltai),
	};
	return qtrun(&p);
}

#ifndef COMBINED_BINARY
int main() {
	print_header();
	return tunmain();
}
#endif


int qtrun(struct qtproto* p) {
        struct qtsession session;
        session.protocol = *p;

        init_udp(&session);
        int sfd = session.fd_socket;
        if (sfd == -1) return -1;

        session.fd_dev = init_tuntap();
        int ttfd = session.fd_dev;
        if (ttfd == -1) return -1;

        char protocol_data[p->protocol_data_size];
        session.protocol_data = &protocol_data;
        if (p->init) p->init(&session);

        fprintf(stderr, "The tunnel is now operational!\n");

        struct pollfd fds[2];
        fds[0].fd = ttfd;
        fds[0].events = POLLIN;
        fds[1].fd = sfd;
        fds[1].events = POLLIN;

        struct sockaddr_in recvaddr;

        char buffer_raw_a[p->buffersize_raw];
        char buffer_enc_a[p->buffersize_enc];
        char* buffer_raw = buffer_raw_a;
        char* buffer_enc = buffer_enc_a;

        while (1) {
                int len = poll(fds, 2, -1);
                if (len < 0) return errorexitp("poll error");
                else if (fds[0].revents & (POLLERR | POLLHUP | POLLNVAL)) return errorexit("poll error on tap device");
                else if (fds[1].revents & (POLLHUP | POLLNVAL)) return errorexit("poll error on udp socket");
                if (fds[0].revents & POLLIN) {
                        len = read(ttfd, buffer_raw + p->offset_raw, p->buffersize_raw);
                        if (session.remote_float == 0 || session.remote_float == 2) {
                                len = p->encode(&session, buffer_raw, buffer_enc, len);
                                if (len < 0) return len;
                                if (session.remote_float == 0) {
                                        write(sfd, buffer_enc + p->offset_enc, len);
                                } else {
                                        sendto(sfd, buffer_enc + p->offset_enc, len, 0, (struct sockaddr*)&session.remote_addr, sizeof(session.remote_addr));
                                }
                        }
                }
                if (fds[1].revents & POLLIN) {
                        socklen_t recvaddr_len = sizeof(recvaddr);
                        if (session.remote_float == 0) {
                                len = read(sfd, buffer_enc + p->offset_enc, p->buffersize_enc);
                        } else {
                                len = recvfrom(sfd, buffer_enc + p->offset_enc, p->buffersize_enc, 0, (struct sockaddr*)&recvaddr, &recvaddr_len);
                        }
                        if (len < 0) {
                                int out;
                                len = 4;
                                getsockopt(sfd, SOL_SOCKET, SO_ERROR, &out, &len);
                                fprintf(stderr, "End of file on udp socket");
                        } else {
                                len = p->decode(&session, buffer_enc, buffer_raw, len);
                                if (len != 0 && session.remote_float != 0 && (session.remote_addr.sin_addr.s_addr != recvaddr.sin_addr.s_addr || session.remo$
                                        fprintf(stderr, "Remote endpoint has changed to %08X:%d\n", recvaddr.sin_addr, ntohs(recvaddr.sin_port));
                                        session.remote_addr = recvaddr;
                                        session.remote_float = 2;
                                }
                                if (len < 0) return len;
                                write(ttfd, buffer_raw + p->offset_raw, len);
                        }
                }
        }
}
